import unittest

from .unet import UNet


class TestUNet(unittest.TestCase):
    def test_construct(self):
        unet = UNet(name="unet50_fmow_rgb")
        n_param = unet.num_params()

        self.assertEqual(n_param, 32541792)

        self.assertTrue(len(unet.describe()))
        self.assertEqual(unet.name, "unet50_fmow_rgb")

    @unittest.skip("Big download need to find a better unittest")
    def test_load_full(self):
        unet = UNet(name="unet50_fmow_full")
        n_param = unet.num_params()

        self.assertEqual(n_param, 32557472)
        self.assertTrue(len(unet.describe()))
        self.assertEqual(unet.name, "unet50_fmow_full")

        unet.load_weights(
            encoder_weights="unet50_fmow_full", decoder_weights="unet50_fmow_full"
        )

    @unittest.skip("Big download need to find a better unittest")
    def test_load_rgb(self):
        unet = UNet(name="unet50_fmow_rgb")
        n_param = unet.num_params()

        self.assertEqual(n_param, 32541792)
        self.assertTrue(len(unet.describe()))
        self.assertEqual(unet.name, "unet50_fmow_rgb")

        unet.load_weights(
            encoder_weights="unet50_fmow_rgb", decoder_weights="unet50_fmow_rgb"
        )

    def test_fails(self):
        with self.assertRaises(Exception):
            _ = UNet(name="not_a_network")


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from .preprocess import get_preprocessing_fn


class TestPreprocess(unittest.TestCase):
    def test_preprocess_fn(self):
        fn = get_preprocessing_fn("unet", "fmow_rgb")

        data = np.ones((8, 224, 224, 3))
        processed = fn(data)
        self.assertAlmostEqual(processed.sum(), 2937294.901659388)

        data = np.ones((1, 224, 224, 3))
        processed = fn(data)
        self.assertAlmostEqual(processed.sum(), 2937294.901659388 / 8)

    def test_preprocess_fn_full(self):
        fn = get_preprocessing_fn("unet", "fmow_full")

        data = np.ones((8, 224, 224, 8))
        processed = fn(data)
        self.assertAlmostEqual(processed.sum(), -9124595.297783388)

        data = np.ones((1, 224, 224, 8))
        processed = fn(data)
        self.assertAlmostEqual(processed.sum(), -9124595.297783388 / 8)


if __name__ == "__main__":
    unittest.main()



import functools
from typing import Callable

import numpy as np

from .settings import DatasetSettings, ModelSettings


def get_dataset_settings(dataset: str) -> DatasetSettings:
    if dataset == "fmow_rgb":
        return DatasetSettings(
            mean=np.array([0.485, 0.456, 0.406]),
            std=np.array([0.229, 0.224, 0.225]),
        )
    elif dataset == "fmow_full":
        return DatasetSettings(
            mean=np.array(
                [349.23, 339.76, 378.58, 418.42, 275.86, 431.82, 495.65, 435.05]
            ),
            std=np.array(
                [78.67, 105.54, 142.05, 177.00, 132.29, 151.65, 194.00, 166.27]
            ),
        )
    else:
        raise ValueError("Invalid dataset type.")


def get_model_settings(model: str) -> ModelSettings:
    if model == "unet":
        return ModelSettings(name="unet")
    else:
        raise ValueError("Invalid dataset type.")


def _preprocess_fn(
    x: np.ndarray, model: ModelSettings, dataset: DatasetSettings
) -> np.ndarray:
    if dataset.mean is not None:
        x = x - dataset.mean

    if dataset.std is not None:
        x = x / dataset.std

    return x


def get_preprocessing_fn(model: str, dataset: str) -> Callable:
    """Get a preprocessing function for a given model and dataset. For a full list of preprocessing functions available, see the documentation.

    Args:
        model: Which type of model to preprocess for, e.g. unet.
        dataset: Which dataset to expect for preprocessing, e.g. fmow_rgb.

    Returns:
        fn: A function that can be applied to an input array.
    """
    model_settings = get_model_settings(model)
    dataset_settings = get_dataset_settings(dataset)

    return functools.partial(
        _preprocess_fn, model=model_settings, dataset=dataset_settings
    )



import argparse
import os
import glob

import tqdm
from joblib import Parallel, delayed

import numpy as np

from utils import read_tile, mask_from_tile


def main(args):
    data_path = args.data_path

    # Process code for paralleization
    def process_file(file):
        # Get the filename for the mask data
        mask_path = os.path.join(data_path, "mask")
        filename = os.path.basename(file)
        keyname = filename.split(".")[0]
        label_name = keyname.replace("8band_", "Mask_") + ".npy"
        write_path = os.path.join(mask_path, label_name)

        # Don't process existing files
        if os.path.exists(write_path):
            return

        try:
            data, transform = read_tile(file)
            mask = mask_from_tile(file, data.shape, transform)
        except:
            print(f"Failed to create mask for {file}")
            return

        np.save(write_path, mask)

    mask_path = os.path.join(data_path, "mask")
    if not os.path.exists(mask_path):
        os.makedirs(mask_path)

    # Run the job in parallel
    files = glob.glob(os.path.join(data_path, "8band/*.tif"))
    Parallel(n_jobs=16)(delayed(process_file)(file) for file in tqdm.tqdm(files))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Create a set of masks from geojson labels."
    )

    parser.add_argument(
        "data_path", type=str, help="Path to the root of the SpaceNet-1 data."
    )
    args = parser.parse_args()

    main(args)



"""Only use loguru if it's present in the environment."""
from typing import Any

logger: Any = None
try:
    from loguru import logger as loguru_logger

    logger = loguru_logger
except:
    import logging

    logger = logging.getLogger("moonshine")



import dataclasses

import numpy as np


@dataclasses.dataclass
class DatasetSettings(object):
    mean: np.ndarray
    std: np.ndarray


@dataclasses.dataclass
class ModelSettings(object):
    name: str



import os

import geojson
import rasterio

import numpy as np

from shapely.geometry import shape, Point


def one_hot_3d(array, num_classes):
    # Get the shape of the input array
    shape = array.shape

    # Create an empty 3D array with the required shape
    one_hot_array = np.zeros((shape[0], shape[1], num_classes))

    # Fill the one-hot array
    for i in range(shape[0]):
        for j in range(shape[1]):
            class_index = array[i, j]
            one_hot_array[i, j, class_index] = 1

    return one_hot_array


def tile_to_rgb(tile):
    rgb = tile[[4, 2, 1], :, :]
    rgb[rgb >= 2000] = 2000
    rgb = rgb / 2000.0
    rgb = np.rollaxis(rgb, 0, 3)

    return rgb


def label_from_8band(path, label_type="geojson"):
    if label_type not in ("geojson", "mask"):
        raise

    # Get the label folder
    folder = os.path.dirname(path)
    up = "/".join(path.split("/")[0:-2])
    label_path = os.path.join(up, label_type)

    # Get the filename
    filename = os.path.basename(path)
    keyname = filename.split(".")[0]

    if label_type == "geojson":
        label_name = keyname.replace("8band_", "Geo_") + ".geojson"
    else:
        label_name = keyname.replace("8band_", "Mask_") + ".npy"

    # Combine them
    return os.path.join(label_path, label_name)


def mask_from_geojson(label, mask_shape, transform):
    M, N, C = mask_shape
    mask = np.zeros((M, N))
    if len(label.features) == 0:
        return mask

    rows, cols = np.meshgrid(np.arange(M), np.arange(N))
    xs, ys = rasterio.transform.xy(transform, rows, cols)
    lats = np.stack(xs, axis=1)
    lons = np.stack(ys, axis=1)
    points = np.stack((lats, lons), axis=2)

    def fn(pt):
        point = Point(pt)
        for feature in label.features:
            polygon = shape(feature["geometry"])
            if polygon.contains(point):
                return 1
        return 0

    mask = np.apply_along_axis(fn, 2, points)
    return mask


def mask_from_tile(path, shape, transform):
    # Get the label
    label_path = label_from_8band(path)
    with open(label_path, "r") as f:
        label_data = geojson.load(f)

    # Turn the json label into another image
    return mask_from_geojson(label_data, shape, transform)


def read_tile(path):
    # Read the data
    with rasterio.open(path, "r") as f:
        data = f.read().astype(float)
        data = np.rollaxis(data, 0, 3)
        transform = f.transform

    return data, transform



# Configuration file for the Sphinx documentation builder.

# -- For some crazy reason this is a requirement to import the package
from typing import Any, List, Tuple, Type

import types
import os
from re import A
import sys
import shutil
import glob

# import torch

import moonshine.models
import moonshine.preprocessing

sys.path.insert(0, os.path.abspath("../.."))
project_root = os.path.abspath("../..")
print("Running doc generation, project root at {}".format(project_root))

# -- Project information
project = "Moonshine"
copyright = "2023, Moonshine Labs"
author = "Moonshine Labs"

release = "0.1.5"
version = "0.1.5"

# -- Theme options
html_static_path = ["_static"]
html_title = " Moonshine"

# Favicon
html_favicon = "https://moonshine-assets.s3.us-west-2.amazonaws.com/favicon.ico"

# Customize CSS
html_css_files = ["css/custom.css"]

# -- Options for HTML output
html_theme = "furo"
html_theme_options = {
    "light_logo": "logo-light-mode.png",
    "dark_logo": "logo-light-mode.png",
    "light_css_variables": {
        "color-brand-primary": "#373737",
        "color-brand-content": "#373737",
    },
    "dark_css_variables": {
        "color-brand-primary": "#f9f9f9",
        "color-brand-content": "#f9f9f9",
    },
}

# -- General configuration
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "myst_parser",
    "nbsphinx",
]
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

pygments_style = "manni"
pygments_dark_style = "monokai"

autosummary_generate = True

nbsphinx_execute = "never"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = [".rst", ".md"]

# -- Options for EPUB output
epub_show_urls = "footnote"

# -- Move the example notebooks into the doc source tree
print("Copy example notebooks into docs/_examples")
examples_root = os.path.join(project_root, "docs/source/_examples")
shutil.rmtree(examples_root, ignore_errors=True)
os.makedirs(examples_root)
pynb_files = glob.glob(os.path.join(project_root, "examples/**/*.ipynb"))

for f in pynb_files:
    new_path = os.path.join(examples_root, os.path.basename(f))
    shutil.copyfile(f, new_path)


def _auto_rst_for_module(module: types.ModuleType, exclude_members: List[Any]) -> str:
    """Generate the content of an rst file documenting a module.
    Includes the module docstring, followed by tables for the functions,
    classes, and exceptions
    Args:
        module: The module object to document
        exclude_members: A list of Python objects to exclude from the
            documentation. Providing objects that are not imported in
            ``module`` are ignored.
    Returns:
        The rst content for the module
    """
    name = module.__name__
    lines = []

    functions: List[Tuple[str, types.FunctionType]] = []
    exceptions: List[Tuple[str, Type[BaseException]]] = []
    classes: List[Tuple[str, Type[object]]] = []
    methods: List[Tuple[str, types.MethodType]] = []
    attributes: List[Tuple[str, object]] = []

    # add title and module docstring
    lines.append(f"{name}")
    lines.append(f'{"=" * len(name)}\n')
    lines.append(f".. automodule:: {name}\n")

    # set prefix so that we can use short names in the autosummaries
    lines.append(f".. currentmodule:: {name}")

    try:
        all_members = list(module.__all__)
    except AttributeError:
        all_members = list(vars(module).keys())

    for item_name, val in vars(module).items():
        if val in exclude_members:
            continue

        if item_name.startswith("_"):
            # Skip private members
            continue

        if item_name not in all_members:
            # Skip members not in `__all__``
            continue

        if isinstance(val, types.ModuleType):
            # Skip modules; those are documented by autosummary
            continue

        if isinstance(val, types.FunctionType):
            functions.append((item_name, val))
        elif isinstance(val, types.MethodType):
            methods.append((item_name, val))
        elif isinstance(val, type) and issubclass(val, BaseException):
            exceptions.append((item_name, val))
        elif isinstance(val, type):
            assert issubclass(val, object)
            classes.append((item_name, val))
        else:
            attributes.append((item_name, val))
            continue

    # Sort by the reimported name
    functions.sort(key=lambda x: x[0])
    exceptions.sort(key=lambda x: x[0])
    classes.sort(key=lambda x: x[0])
    attributes.sort(key=lambda x: x[0])

    for category, category_name in (
        (functions, "Functions"),
        (classes, "Classes"),
        (exceptions, "Exceptions"),
    ):
        sphinx_lines = []
        for item_name, _ in category:
            sphinx_lines.append(f"      {item_name}")
        if len(sphinx_lines) > 0:
            lines.append(f"\n.. rubric:: {category_name}\n")
            lines.append(".. autosummary::")
            lines.append("      :toctree: generated")
            lines.append("      :nosignatures:")
            if category_name in ("Classes"):
                lines.append("      :template: classtemplate.rst")
            elif category_name == "Functions":
                lines.append("      :template: functemplate.rst")
            lines.append("")
            lines.extend(sphinx_lines)
            lines.append("")

    lines.append(".. This file autogenerated by docs/source/conf.py\n")

    return "\n".join(lines)


def _modules_to_rst() -> List[types.ModuleType]:
    """Return the list of modules for which to generate API reference rst files."""
    document_modules: List[types.Module] = [
        moonshine.models,
        moonshine.preprocessing,
    ]
    exclude_modules: List[types.Module] = [moonshine]
    for name in moonshine.__dict__:
        obj = moonshine.__dict__[name]
        if isinstance(obj, types.ModuleType) and obj not in exclude_modules:
            document_modules.append(obj)

    return document_modules


def _generate_rst_files_for_modules() -> None:
    """Generate .rst files for each module to include in the API reference.
    These files contain the module docstring followed by tables listing all the functions, classes,
    etc.
    """
    docs_dir = os.path.abspath(os.path.dirname(__file__))
    module_rst_save_dir = os.path.join(docs_dir, "api_reference")
    # gather up modules to generate rst files for
    document_modules = _modules_to_rst()

    # rip out types that are duplicated in top-level moonshine module
    moonshine_imported_types = []
    for name in moonshine.__all__:
        obj = moonshine.__dict__[name]
        if not isinstance(obj, types.ModuleType):
            moonshine_imported_types.append(obj)

    document_modules = sorted(document_modules, key=lambda x: x.__name__)
    os.makedirs(module_rst_save_dir, exist_ok=True)
    for module in document_modules:
        saveas = os.path.join(module_rst_save_dir, module.__name__ + ".rst")
        print(f"Generating rst file {saveas} for module: {module.__name__}")

        # avoid duplicate entries in docs. We add torch's _LRScheduler to
        # types, so we get a ``WARNING: duplicate object description`` if we
        # don't exclude it
        # exclude_members = [torch.optim.lr_scheduler._LRScheduler]
        exclude_members = []
        if module is not moonshine:
            exclude_members += moonshine_imported_types

        content = _auto_rst_for_module(module, exclude_members=exclude_members)

        with open(saveas, "w") as f:
            f.write(content)


_generate_rst_files_for_modules()



from json import encoder

"""UNet models for segmentation and classification."""
from typing import Optional, Sequence

import segmentation_models_pytorch as smp

from .base import MoonshineModel
from .model_parameters import model_params


class UNet(MoonshineModel):
    """A basic UNet model, implemented under the hood with segmentation-models-pytorch. The model uses a ResNet backbone and skip connections, as in the original paper at https://arxiv.org/abs/1505.04597. Some parameters are fixed, since pre-trained weights require a consistant network structure."""

    def _build_model(self) -> smp.Unet:
        assert self.name in model_params.keys(), "Unsupported model type."

        encoder = model_params[self.name]["encoder"]
        channels = model_params[self.name]["input_channels"]
        decoder = model_params[self.name]["decoder"]
        classes = model_params[self.name]["output_channels"]

        return smp.Unet(
            encoder_name=encoder,
            encoder_weights=None,
            decoder_channels=decoder,
            in_channels=channels,
            classes=classes,
        )

    def __init__(
        self,
        name,
    ):
        """Create the UNet, without loading the weights.

        Args:
            name: A valid name for the architecture of this model.
        """
        super().__init__(name=name)
        self.unet = self._build_model()

    def forward(self, x):
        """Run the forward pass of this model."""
        return self.unet(x)



# TODO(nharada): Make this dataclasses instead for better type checking

model_params = {
    "unet50_fmow_rgb": {
        "weight_url": "https://moonshine-prod-models.s3.us-west-2.amazonaws.com/moonshine/unet/unet50_fmow_3chan_step268380.pt",
        "encoder": "resnet50",
        "decoder": (256, 128, 64, 32, 32),
        "input_channels": 3,
        "output_channels": 32,
    },
    "unet50_fmow_full": {
        "weight_url": "https://moonshine-prod-models.s3.us-west-2.amazonaws.com/moonshine/unet/unet50_fmow_allchan_step284000.pt",
        "encoder": "resnet50",
        "decoder": (256, 128, 64, 32, 32),
        "input_channels": 8,
        "output_channels": 32,
    },
}



"""The ABC for Moonshine models."""
import abc
import io
import logging
from json import decoder
from multiprocessing.sharedctypes import Value
from typing import Any, Optional

import torch
import torch.nn as nn
from smart_open import open
from torch.utils.model_zoo import load_url

from .logging import logger
from .model_parameters import model_params


class MoonshineModel(nn.Module, abc.ABC):
    """The base class of all Moonshine released models."""

    def __init__(self, name: str):
        """Create the moonshine base model.

        Args:
            name: A valid name for the architecture of this model.
        """
        super().__init__()

        self.name = name

    def _download_state_dict(self, location: str) -> dict[str, Any]:
        device = "cuda" if torch.cuda.is_available() else "cpu"

        if "s3://" in location:
            with open(location, "rb") as f:
                buffer = io.BytesIO(f.read())
                return torch.load(buffer)
        else:
            return load_url(location, map_location=torch.device(device))

    def _load_state(
        self, encoder_weights: Optional[str], decoder_weights: Optional[str]
    ):
        new_dict = {}
        # Load the encoder weights
        if encoder_weights:
            state_dict = self._download_state_dict(encoder_weights)
            for k, v in state_dict.items():
                if "unet" in k and "encode" in k:
                    new_key = k.replace("model.", "")
                    new_dict[new_key] = v

        # Load the decoder weights
        if decoder_weights:
            state_dict = self._download_state_dict(decoder_weights)
            for k, v in state_dict.items():
                if "unet" in k and "decode" in k:
                    new_key = k.replace("model.", "")
                    new_dict[new_key] = v

        self.load_state_dict(new_dict, strict=False)

    def load_weights(
        self,
        encoder_weights: Optional[str],
        decoder_weights: Optional[str],
    ):
        """Load external weights for this model. Can be either a path or a
        named set.

        Args:
            encoder_weights: Either a path to .pt weights or a valid model name. If None, will not load encoder weights.
            decoder_weights: Either a path to .pt weights or a valid model name. If None, will not load decoder weights.
        """
        if not encoder_weights and not decoder_weights:
            raise ValueError(
                "Didn't get any weights at all, need either encoder or decoder."
            )

        encoder_url = encoder_weights
        if encoder_weights:
            if encoder_weights in model_params.keys():
                encoder_url = str(model_params[encoder_weights]["weight_url"])
            logger.info(f"Trying to load encoder weights from {encoder_url}")

        decoder_url = decoder_weights
        if decoder_weights:
            if decoder_weights in model_params.keys():
                decoder_url = str(model_params[decoder_weights]["weight_url"])
            logger.info(f"Trying to load decoder weights from {decoder_url}")

        self._load_state(encoder_url, decoder_url)

    def num_params(self):
        trainable_params = 0
        for p in self.parameters():
            if p.requires_grad:
                trainable_params += p.numel()

        return trainable_params

    def describe(self):
        description = f"Moonshine Model: {self.name}"
        description += f"   Number of trainable parameters: {self.num_params()}"

        return "\n".join(description)
